overlay elpi https://github.com/SkySkimmer/coq-elpi reduction 17453

overlay equations https://github.com/SkySkimmer/Coq-Equations reduction 17453

overlay unicoq https://github.com/SkySkimmer/unicoq reduction 17453

overlay mtac2 https://github.com/SkySkimmer/Mtac2 reduction 17453

overlay relation_algebra https://github.com/SkySkimmer/relation-algebra reduction 17453

overlay serapi https://github.com/SkySkimmer/coq-serapi reduction 17453

overlay smtcoq https://github.com/SkySkimmer/smtcoq reduction 17453
