overlay elpi https://github.com/SkySkimmer/coq-elpi warn-api 17585

overlay equations https://github.com/SkySkimmer/Coq-Equations warn-api 17585

overlay metacoq https://github.com/SkySkimmer/metacoq warn-api 17585

overlay quickchick https://github.com/SkySkimmer/QuickChick warn-api 17585

overlay smtcoq https://github.com/SkySkimmer/smtcoq warn-api 17585
