val lu : (int * int) list
val ll : (int * int) list
val lt : (int * int) list
val mn : (int * int) list
val mc : (int * int) list
val me : (int * int) list
val nd : (int * int) list
val nl : (int * int) list
val no : (int * int) list
val zs : (int * int) list
val zl : (int * int) list
val zp : (int * int) list
val cc : (int * int) list
val cf : (int * int) list
val cs : (int * int) list
val co : (int * int) list
val cn : (int * int) list
val lm : (int * int) list
val lo : (int * int) list
val pc : (int * int) list
val pd : (int * int) list
val ps : (int * int) list
val pe : (int * int) list
val pi : (int * int) list
val pf : (int * int) list
val po : (int * int) list
val sm : (int * int) list
val sc : (int * int) list
val sk : (int * int) list
val so : (int * int) list
val to_lower : ((int * int) * [> `Abs of int | `Delta of int ]) list
