overlay coq_lsp https://github.com/SkySkimmer/coq-lsp summary-local 17605

overlay elpi https://github.com/SkySkimmer/coq-elpi summary-local 17605

overlay lean_importer https://github.com/SkySkimmer/coq-lean-import summary-local 17605
