Require Import Coq.Lists.List.
Require Import Coq.ZArith.ZArith.
Goal True.
  exact I.
Qed.
Goal True.
  exact I.
Qed.
Goal True.
  exact I.
Qed.
Goal True.
  exact I.
Qed.
Goal True.
  exact I.
Qed.
Goal List.repeat Z.div_eucl 5 = List.repeat Z.div_eucl 5.
  vm_compute; reflexivity.
Qed.
