val logic : string
val symbol : string
val fraction : string
val letter : string
val greek_letter : string
val asciiart : string
val equivalence : string
val order : string
val circle : string
val square : string
val triangle : string
val arrow : string
val set : string
val math : string
val space : string
val delimiter : string
val miscellanea : string
val bindings_set_1 : (string list * string * string list) list
val bindings_set_2 : (string * string) list
val priorities : (string * int) list
val filename : string
