Require Import Ltac2.Ltac2.

Check $x.

Check $preterm:x.

Check $ x.

Check $ preterm:x.

Check $ preterm : x.

Check $preterm : x.

Check fun x => $preterm : x.

Check $preterm: x.

Check fun x => $preterm: x.

Check $ preterm :x.
