overlay elpi https://github.com/SkySkimmer/coq-elpi depr-versions 17664

overlay equations https://github.com/SkySkimmer/Coq-Equations depr-versions 17664

overlay metacoq https://github.com/SkySkimmer/metacoq depr-versions 17664

overlay itauto https://github.com/SkySkimmer/https://gitlab.inria.fr/fbesson/itauto depr-versions 17664

overlay lean_importer https://github.com/SkySkimmer/coq-lean-import depr-versions 17664

overlay unicoq https://github.com/SkySkimmer/unicoq depr-versions 17664

overlay paramcoq https://github.com/SkySkimmer/paramcoq depr-versions 17664

overlay serapi https://github.com/SkySkimmer/coq-serapi depr-versions 17664

overlay tactician https://github.com/SkySkimmer/coq-tactician depr-versions 17664

overlay coqhammer https://github.com/SkySkimmer/coqhammer depr-versions 17664
