Print Tables.
Print ML Path.
Print ML Modules.
Print LoadPath.
Print Graph.
Print Coercions.
Print Classes.
Print nat.
Print Term O.
Print All.
Print Grammar constr.
Inspect 10.

Section A.
Coercion f (x : nat) : Prop := True.
Print Coercion Paths nat Sortclass.

Print Section A.

End A.
