overlay coqhammer https://github.com/SkySkimmer/coqhammer app-head 17633

overlay equations https://github.com/SkySkimmer/Coq-Equations app-head 17633

overlay metacoq https://github.com/SkySkimmer/metacoq app-head 17633

overlay lean_importer https://github.com/SkySkimmer/coq-lean-import app-head 17633

overlay unicoq https://github.com/SkySkimmer/unicoq app-head 17633

overlay mtac2 https://github.com/SkySkimmer/Mtac2 app-head 17633

overlay quickchick https://github.com/SkySkimmer/QuickChick app-head 17633

overlay reduction_effects https://github.com/SkySkimmer/reduction-effects app-head 17633

overlay smtcoq https://github.com/SkySkimmer/smtcoq app-head 17633
