overlay elpi https://github.com/proux01/coq-elpi coq_17754 17754

overlay serapi https://github.com/proux01/coq-serapi coq_17754 17754

overlay coq_lsp https://github.com/proux01/coq-lsp coq_17754 17754
