overlay elpi https://github.com/SkySkimmer/coq-elpi deprecate-definitions 16890

overlay equations https://github.com/SkySkimmer/Coq-Equations deprecate-definitions 16890

overlay mtac2 https://github.com/SkySkimmer/Mtac2 deprecate-definitions 16890

overlay metacoq https://github.com/SkySkimmer/metacoq deprecate-definitions 16890
